﻿using System;
using System.Threading;
using System.Threading.Tasks;

namespace TestLibrary
{
    /// <summary>
    /// 취소 확장
    /// </summary>
    public static class CancellationExtension
    {
        //////////////////////////////////////////////////////////////////////////////////////////////////// Method
        ////////////////////////////////////////////////////////////////////////////////////////// Static
        //////////////////////////////////////////////////////////////////////////////// Public

        #region 취소시 태스크 구하기 - WhenCanceled(cancellationToken)

        /// <summary>
        /// 취소시 태스크 구하기
        /// </summary>
        /// <param name="cancellationToken">취소 토큰</param>
        /// <returns>취소시 태스크</returns>
        public static Task WhenCanceled(this CancellationToken cancellationToken)
        {
            TaskCompletionSource<int> source = new TaskCompletionSource<int>();

            CancellationTokenRegistration registration = default(CancellationTokenRegistration);

            registration = cancellationToken.Register
            (
                o =>
                {
                    ((TaskCompletionSource<int>)o).TrySetCanceled();

                    registration.Dispose();
                },
                source
            );

            return source.Task;
        }

        #endregion
        #region 취소 토큰을 갖는 태스크 구하기 - WithCancellationToken(task, cancellationToken)

        /// <summary>
        /// 취소 토큰을 갖는 태스크 구하기
        /// </summary>
        /// <param name="task">태스크</param>
        /// <param name="cancellationToken">취소 토큰</param>
        /// <returns></returns>
        public static async Task WithCancellationToken(this Task task, CancellationToken cancellationToken)
        {
            await await Task.WhenAny(task, cancellationToken.WhenCanceled());
        }

        #endregion
        #region 취소 토큰을 갖는 태스크 구하기 - WithCancellationToken<T>(task, cancellationToken)

        /// <summary>
        /// 취소 토큰을 갖는 태스크 구하기
        /// </summary>
        /// <typeparam name="T">타입</typeparam>
        /// <param name="task">태스크</param>
        /// <param name="cancellationToken">취소 토큰</param>
        /// <returns>태스크</returns>
        public static async Task<T> WithCancellationToken<T>(this Task<T> task, CancellationToken cancellationToken)
        {
            Task firstTaskToFinish = await Task.WhenAny(task, cancellationToken.WhenCanceled());

            if(firstTaskToFinish == task)
            {
                return await task;
            }

            await firstTaskToFinish;

            throw new OperationCanceledException(cancellationToken);
        }

        #endregion
    }
}