using System.Collections.Generic;
using System.IO;
using System.Linq;
using System.Threading.Tasks;

namespace TestLibrary.Decoding
{
    /// <summary>
    /// GIF  Ʈ
    /// </summary>
    public class GIFDataStream
    {
        //////////////////////////////////////////////////////////////////////////////////////////////////// Property
        ////////////////////////////////////////////////////////////////////////////////////////// Public

        #region  - Header

        /// <summary>
        /// 
        /// </summary>
        public GIFHeader Header { get; private set; }

        #endregion
        #region   ̺ - GlobalColorTable

        /// <summary>
        ///   ̺
        /// </summary>
        public GIFColor[] GlobalColorTable { get; set; }

        #endregion
        #region GIF  Ʈ - FrameList

        /// <summary>
        /// GIF  Ʈ
        /// </summary>
        public IList<GIFFrame> FrameList { get; set; }

        #endregion
        #region GIF Ȯ Ʈ - ExtensionList

        /// <summary>
        /// GIF Ȯ Ʈ
        /// </summary>
        public IList<GIFExtension> ExtensionList { get; set; }

        #endregion
        #region ݺ īƮ - RepeatCount

        /// <summary>
        /// ݺ īƮ
        /// </summary>
        public ushort RepeatCount { get; set; }

        #endregion

        //////////////////////////////////////////////////////////////////////////////////////////////////// Constructor
        ////////////////////////////////////////////////////////////////////////////////////////// Private

        #region  - GIFDataStream()

        /// <summary>
        /// 
        /// </summary>
        private GIFDataStream()
        {
        }

        #endregion

        //////////////////////////////////////////////////////////////////////////////////////////////////// Method
        ////////////////////////////////////////////////////////////////////////////////////////// Static
        //////////////////////////////////////////////////////////////////////////////// Public

        #region б (񵿱) - ReadAsync(stream)

        /// <summary>
        /// б (񵿱)
        /// </summary>
        /// <param name="stream">Ʈ</param>
        /// <returns>GIF  Ʈ ½ũ</returns>
        public static async Task<GIFDataStream> ReadAsync(Stream stream)
        {
            GIFDataStream dataStream = new GIFDataStream();

            await dataStream.ReadInternalAsync(stream).ConfigureAwait(false);

            return dataStream;
        }

        #endregion

        ////////////////////////////////////////////////////////////////////////////////////////// Instance
        //////////////////////////////////////////////////////////////////////////////// Private

        #region  б (񵿱) - ReadFramesAsync(stream)

        /// <summary>
        ///  б (񵿱)
        /// </summary>
        /// <param name="stream">Ʈ</param>
        /// <returns>½ũ</returns>
        private async Task ReadFramesAsync(Stream stream)
        {
            List<GIFFrame>     frameList            = new List<GIFFrame>();
            List<GIFExtension> controlExtensionList = new List<GIFExtension>();
            List<GIFExtension> specialExtensionList = new List<GIFExtension>();

            while(true)
            {
                try
                {
                    GIFBlock block = await GIFBlock.ReadAsync(stream, controlExtensionList).ConfigureAwait(false);

                    if(block.Kind == GIFBlockKind.GraphicRendering)
                    {
                        controlExtensionList = new List<GIFExtension>();
                    }

                    if(block is GIFFrame frame)
                    {
                        frameList.Add(frame);
                    }
                    else if(block is GIFExtension extension)
                    {
                        switch(extension.Kind)
                        {
                            case GIFBlockKind.Control :

                                controlExtensionList.Add(extension);

                                break;

                            case GIFBlockKind.SpecialPurpose :

                                specialExtensionList.Add(extension);

                                break;
                        }
                    }
                    else if(block is GIFTrailer)
                    {
                        break;
                    }
                }
                catch(UnknownBlockTypeException) when (frameList.Count > 0)
                {
                    break;
                }
            }

            this.FrameList     = frameList.AsReadOnly();
            this.ExtensionList = specialExtensionList.AsReadOnly();
        }

        #endregion
        #region б (ο, 񵿱) - ReadInternalAsync(stream)

        /// <summary>
        /// б (ο, 񵿱)
        /// </summary>
        /// <param name="stream">Ʈ</param>
        /// <returns>½ũ</returns>
        private async Task ReadInternalAsync(Stream stream)
        {
            Header = await GIFHeader.ReadAsync(stream).ConfigureAwait(false);

            if(Header.LogicalScreenDescriptor.HasGlobalColorTable)
            {
                GlobalColorTable = await GIFHelper.ReadColorTableAsync
                (
                    stream,
                    Header.LogicalScreenDescriptor.GlobalColorTableSize
                ).ConfigureAwait(false);
            }

            await ReadFramesAsync(stream).ConfigureAwait(false);

            GIFApplicationExtension extension = ExtensionList.OfType<GIFApplicationExtension>()
                .FirstOrDefault(GIFHelper.IsNetscapeExtension);

            RepeatCount = extension != null ? GIFHelper.GetRepeatCount(extension) : (ushort)1;
        }

        #endregion
    }
}