using System.Collections.Generic;
using System.IO;
using System.Linq;
using System.Threading.Tasks;

namespace TestLibrary.Decoding
{
    /// <summary>
    /// GIF 
    /// </summary>
    public class GIFFrame : GIFBlock
    {
        //////////////////////////////////////////////////////////////////////////////////////////////////// Field
        ////////////////////////////////////////////////////////////////////////////////////////// Public

        #region Field

        /// <summary>
        /// ̹ и
        /// </summary>
        public const int ImageSeparator = 0x2C;

        #endregion

        //////////////////////////////////////////////////////////////////////////////////////////////////// Property
        ////////////////////////////////////////////////////////////////////////////////////////// Public

        #region GIF ̹  - Descriptor

        /// <summary>
        /// GIF ̹ 
        /// </summary>
        public GIFImageDescriptor Descriptor { get; private set; }

        #endregion
        #region   ̺ - LocalColorTable

        /// <summary>
        ///   ̺
        /// </summary>
        public GIFColor[] LocalColorTable { get; private set; }

        #endregion
        #region GIF Ȯ Ʈ - ExtensionList

        /// <summary>
        /// GIF Ȯ Ʈ
        /// </summary>
        public IList<GIFExtension> ExtensionList { get; private set; }

        #endregion
        #region GIF ̹  - ImageData

        /// <summary>
        /// GIF ̹ 
        /// </summary>
        public GIFImageData ImageData { get; private set; }

        #endregion
        #region GIF ׷ Ʈ Ȯ - GraphicControl

        /// <summary>
        /// GIF ׷ Ʈ Ȯ
        /// </summary>
        public GIFGraphicControlExtension GraphicControl { get; set; }

        #endregion
        #region GIF   - Kind

        /// <summary>
        /// GIF  
        /// </summary>
        public override GIFBlockKind Kind
        {
            get
            {
                return GIFBlockKind.GraphicRendering;
            }
        }

        #endregion

        //////////////////////////////////////////////////////////////////////////////////////////////////// Constructor
        ////////////////////////////////////////////////////////////////////////////////////////// Private

        #region  - GIFFrame()

        /// <summary>
        /// 
        /// </summary>
        private GIFFrame()
        {
        }

        #endregion

        //////////////////////////////////////////////////////////////////////////////////////////////////// Method
        ////////////////////////////////////////////////////////////////////////////////////////// Static
        //////////////////////////////////////////////////////////////////////////////// Public

        #region б (񵿱) - ReadAsync(stream, extensionEnumerable)

        /// <summary>
        /// б (񵿱)
        /// </summary>
        /// <param name="stream">Ʈ</param>
        /// <param name="extensionEnumerable">GIF Ȯ  </param>
        /// <returns>GIF  ½ũ</returns>
        public new static async Task<GIFFrame> ReadAsync(Stream stream, IEnumerable<GIFExtension> extensionEnumerable)
        {
            GIFFrame frame = new GIFFrame();

            await frame.ReadInternalAsync(stream, extensionEnumerable).ConfigureAwait(false);

            return frame;
        }

        #endregion

        ////////////////////////////////////////////////////////////////////////////////////////// Instance
        //////////////////////////////////////////////////////////////////////////////// Private

        #region б (ο, 񵿱) - ReadInternalAsync(stream, extensionEnumerable)

        /// <summary>
        /// б (ο, 񵿱)
        /// </summary>
        /// <param name="stream">Ʈ</param>
        /// <param name="extensionEnumerable">GIF Ȯ  </param>
        /// <returns>½ũ</returns>
        private async Task ReadInternalAsync(Stream stream, IEnumerable<GIFExtension> extensionEnumerable)
        {
            Descriptor = await GIFImageDescriptor.ReadAsync(stream).ConfigureAwait(false);

            if(Descriptor.HasLocalColorTable)
            {
                LocalColorTable = await GIFHelper.ReadColorTableAsync
                (
                    stream,
                    Descriptor.LocalColorTableSize
                ).ConfigureAwait(false);
            }

            ImageData = await GIFImageData.ReadAsync(stream).ConfigureAwait(false);

            ExtensionList  = extensionEnumerable.ToList().AsReadOnly();

            GraphicControl = ExtensionList.OfType<GIFGraphicControlExtension>().FirstOrDefault();
        }

        #endregion
    }
}