using System;
using System.IO;
using System.Text;
using System.Threading;
using System.Threading.Tasks;

using TestLibrary.Extension;

namespace TestLibrary.Decoding
{
    /// <summary>
    /// GIF 
    /// </summary>
    public static class GIFHelper
    {
        //////////////////////////////////////////////////////////////////////////////////////////////////// Method
        ////////////////////////////////////////////////////////////////////////////////////////// Static
        //////////////////////////////////////////////////////////////////////////////// Public

        #region ڿ б (񵿱) - ReadStringAsync(stream, length)

        /// <summary>
        /// ڿ б (񵿱)
        /// </summary>
        /// <param name="stream">Ʈ</param>
        /// <param name="length"></param>
        /// <returns>ڿ ½ũ</returns>
        public static async Task<string> ReadStringAsync(Stream stream, int length)
        {
            byte[] byteArray = new byte[length];

            await stream.ReadAllAsync(byteArray, 0, length).ConfigureAwait(false);

            return GetString(byteArray);
        }

        #endregion
        #region Ʈ   ϱ (񵿱) - CopyDataBlocksToStreamAsync(sourceStream, targetStream, cancellationToken)

        /// <summary>
        /// Ʈ   ϱ (񵿱)
        /// </summary>
        /// <param name="sourceStream">ҽ Ʈ</param>
        /// <param name="targetStream">Ÿ Ʈ</param>
        /// <param name="cancellationToken"> ū</param>
        /// <returns>½ũ</returns>
        public static async Task CopyDataBlocksToStreamAsync(Stream sourceStream, Stream targetStream, CancellationToken cancellationToken = default)
        {
            int length;

            byte[] buffer = new byte[255];

            while((length = await sourceStream.ReadByteAsync(cancellationToken)) > 0)
            {
                await sourceStream.ReadAllAsync(buffer, 0, length, cancellationToken).ConfigureAwait(false);

#if LACKS_STREAM_MEMORY_OVERLOADS

                await targetStream.WriteAsync(buffer, 0, length, cancellationToken);
#else
                await targetStream.WriteAsync(buffer.AsMemory(0, length), cancellationToken);
#endif
            }
        }

        #endregion
        #region   Һϱ (񵿱) - ConsumeDataBlocksAsync(stream, cancellationToken)

        /// <summary>
        ///   Һϱ (񵿱)
        /// </summary>
        /// <param name="stream">Ʈ</param>
        /// <param name="cancellationToken"> ū</param>
        /// <returns>½ũ</returns>
        public static async Task ConsumeDataBlocksAsync(Stream stream, CancellationToken cancellationToken = default)
        {
            await CopyDataBlocksToStreamAsync(stream, Stream.Null, cancellationToken);
        }

        #endregion
        #region   б (񵿱) - ReadDataBlocksAsync(sourceStream, cancellationToken)

        /// <summary>
        ///   б (񵿱)
        /// </summary>
        /// <param name="sourceStream">ҽ Ʈ</param>
        /// <param name="cancellationToken"> ū</param>
        /// <returns>Ʈ 迭 ½ũ</returns>
        public static async Task<byte[]> ReadDataBlocksAsync(Stream sourceStream, CancellationToken cancellationToken = default)
        {
            using MemoryStream targetStream = new MemoryStream();

            await CopyDataBlocksToStreamAsync(sourceStream, targetStream, cancellationToken);

            return targetStream.ToArray();
        }

        #endregion
        #region  ̺ б (񵿱) - ReadColorTableAsync(stream, size)

        /// <summary>
        ///  ̺ б (񵿱)
        /// </summary>
        /// <param name="stream">Ʈ</param>
        /// <param name="size">ũ</param>
        /// <returns>GIF  迭 ½ũ</returns>
        public static async Task<GIFColor[]> ReadColorTableAsync(Stream stream, int size)
        {
            int    length    = 3 * size;
            byte[] byteArray = new byte[length];

            await stream.ReadAllAsync(byteArray, 0, length).ConfigureAwait(false);

            GIFColor[] colorTable = new GIFColor[size];

            for(int i = 0; i < size; i++)
            {
                byte red   = byteArray[3 * i    ];
                byte green = byteArray[3 * i + 1];
                byte blue  = byteArray[3 * i + 2];

                colorTable[i] = new GIFColor(red, green, blue);
            }

            return colorTable;
        }

        #endregion
        #region ݽ Ȯ  ϱ - IsNetscapeExtension(extension)

        /// <summary>
        /// ݽ Ȯ  ϱ
        /// </summary>
        /// <param name="extension">GIF ø̼ Ȯ</param>
        /// <returns>ݽ Ȯ </returns>
        public static bool IsNetscapeExtension(GIFApplicationExtension extension)
        {
            return extension.ApplicationIdentifier == "NETSCAPE" && GetString(extension.AuthenticationCode) == "2.0";
        }

        #endregion
        #region ݺ īƮ ϱ - GetRepeatCount(extension)

        /// <summary>
        /// ݺ īƮ ϱ
        /// </summary>
        /// <param name="extension">GIF ø̼ Ȯ</param>
        /// <returns>ݺ īƮ</returns>
        public static ushort GetRepeatCount(GIFApplicationExtension extension)
        {
            if(extension.Data.Length >= 3)
            {
                return BitConverter.ToUInt16(extension.Data, 1);
            }

            return 1;
        }

        #endregion
        #region     Ÿ   - ThrowUnknownBlockTypeException(blockID)

        /// <summary>
        ///     Ÿ  
        /// </summary>
        /// <param name="blockID"> ID</param>
        /// <returns></returns>
        public static Exception ThrowUnknownBlockTypeException(int blockID)
        {
            return new UnknownBlockTypeException("Unknown block type : 0x" + blockID.ToString("x2"));
        }

        #endregion
        #region    Ȯ Ÿ   - ThrowUnknownExtensionTypeException(extensionLabel)

        /// <summary>
        ///    Ȯ Ÿ  
        /// </summary>
        /// <param name="extensionLabel">Ȯ ̺</param>
        /// <returns></returns>
        public static Exception ThrowUnknownExtensionTypeException(int extensionLabel)
        {
            return new UnknownExtensionTypeException("Unknown extension type : 0x" + extensionLabel.ToString("x2"));
        }

        #endregion
        #region ȿ  ũ   - ThrowInvalidBlockSizeException(blockName, expectedBlockSize, actualBlockSize)

        /// <summary>
        /// ȿ  ũ  
        /// </summary>
        /// <param name="blockName"></param>
        /// <param name="expectedBlockSize">  ũ</param>
        /// <param name="actualBlockSize">  ũ</param>
        /// <returns></returns>
        public static Exception ThrowInvalidBlockSizeException(string blockName, int expectedBlockSize, int actualBlockSize)
        {
            return new InvalidBlockSizeException
            (
                $"Invalid block size for {blockName}. Expected {expectedBlockSize}, but was {actualBlockSize}"
            );
        }

        #endregion
        #region ȿ    - ThrowInvalidSignatureException(signature)

        /// <summary>
        /// ȿ   
        /// </summary>
        /// <param name="signature"></param>
        /// <returns></returns>
        public static Exception ThrowInvalidSignatureException(string signature)
        {
            return new InvalidSignatureException("Invalid file signature : " + signature);
        }

        #endregion
        #region     - ThrowUnsupportedVersionException(version)

        /// <summary>
        ///    
        /// </summary>
        /// <param name="version"></param>
        /// <returns></returns>
        public static Exception ThrowUnsupportedVersionException(string version)
        {
            return new UnsupportedGIFVersionException("Unsupported version : " + version);
        }

        #endregion
        #region ڿ ϱ - GetString(sourceByteArray, index, count)

        /// <summary>
        /// ڿ ϱ
        /// </summary>
        /// <param name="sourceByteArray">ҽ Ʈ 迭</param>
        /// <param name="index">ε</param>
        /// <param name="count">īƮ</param>
        /// <returns>ڿ</returns>
        public static string GetString(byte[] sourceByteArray, int index, int count)
        {
            return Encoding.UTF8.GetString(sourceByteArray, index, count);
        }

        #endregion
        #region ڿ ϱ - GetString(sourceByteArray)

        /// <summary>
        /// ڿ ϱ
        /// </summary>
        /// <param name="sourceByteArray">ҽ Ʈ 迭</param>
        /// <returns>ڿ</returns>
        public static string GetString(byte[] sourceByteArray)
        {
            return GetString(sourceByteArray, 0, sourceByteArray.Length);
        }

        #endregion
    }
}