﻿namespace TestLibrary.Decompression
{
    /// <summary>
    /// 비트 리더
    /// </summary>
    internal class BitReader
    {
        //////////////////////////////////////////////////////////////////////////////////////////////////// Field
        ////////////////////////////////////////////////////////////////////////////////////////// Private

        #region Field

        /// <summary>
        /// 바이트 배열
        /// </summary>
        private readonly byte[] byteArray;

        /// <summary>
        /// 바이트 위치
        /// </summary>
        private int bytePosition = -1;

        /// <summary>
        /// 비트 위치
        /// </summary>
        private int bitPosition;

        /// <summary>
        /// 현재 값
        /// </summary>
        private int currentValue = -1;

        #endregion

        //////////////////////////////////////////////////////////////////////////////////////////////////// Constructor
        ////////////////////////////////////////////////////////////////////////////////////////// Public

        #region 생성자 - BitReader(byteArray)

        /// <summary>
        /// 생성자
        /// </summary>
        /// <param name="byteArray">바이트 배열</param>
        public BitReader(byte[] byteArray)
        {
            this.byteArray = byteArray;
        }

        #endregion

        //////////////////////////////////////////////////////////////////////////////////////////////////// Method
        ////////////////////////////////////////////////////////////////////////////////////////// Public

        #region 비트 읽기 - ReadBits(bitCount)

        /// <summary>
        /// 비트 읽기
        /// </summary>
        /// <param name="bitCount">비트 카운트</param>
        /// <returns>비트</returns>
        public int ReadBits(int bitCount)
        {
            if(this.bytePosition == -1)
            {
                this.bytePosition = 0;
                this.bitPosition  = 0;
                this.currentValue = ReadInteger();
            }
            else if(bitCount > 32 - this.bitPosition)
            {
                int temporaryValue = this.bitPosition >> 3;

                this.bytePosition += temporaryValue;

                this.bitPosition &= 0x07;

                this.currentValue = ReadInteger() >> this.bitPosition;
            }

            int mask  = (1 << bitCount) - 1;
            int value = this.currentValue & mask;

            this.currentValue >>= bitCount;

            this.bitPosition += bitCount;

            return value;
        }

        #endregion

        ////////////////////////////////////////////////////////////////////////////////////////// Private

        #region 정수 읽기 - ReadInteger()

        /// <summary>
        /// 정수 읽기
        /// </summary>
        /// <returns>정수</returns>
        private int ReadInteger()
        {
            int value = 0;

            for(int i = 0; i < 4; i++)
            {
                if(this.bytePosition + i >= this.byteArray.Length)
                {
                    break;
                }

                value |= this.byteArray[this.bytePosition + i] << (i << 3);
            }

            return value;
        }

        #endregion
    }
}