﻿using System.Collections.Generic;
using System.ComponentModel;
using System.Runtime.CompilerServices;
using System.Windows;

namespace TestProject
{
    /// <summary>
    /// 입력 윈도우
    /// </summary>
    public partial class InputWindow : Window, INotifyPropertyChanged
    {
        //////////////////////////////////////////////////////////////////////////////////////////////////// Event
        ////////////////////////////////////////////////////////////////////////////////////////// Public

        #region 속성 변경시 이벤트 - PropertyChanged

        /// <summary>
        /// 속성 변경시 이벤트
        /// </summary>
        public event PropertyChangedEventHandler PropertyChanged;

        #endregion

        //////////////////////////////////////////////////////////////////////////////////////////////////// Field
        ////////////////////////////////////////////////////////////////////////////////////////// Private

        #region Field

        /// <summary>
        /// 메시지
        /// </summary>
        private string message;

        /// <summary>
        /// 텍스트
        /// </summary>
        private string text;

        #endregion

        //////////////////////////////////////////////////////////////////////////////////////////////////// Property
        ////////////////////////////////////////////////////////////////////////////////////////// Public

        #region 메시지 - Message

        /// <summary>
        /// 메시지
        /// </summary>
        public string Message
        {
            get => this.message;
            set => Set(ref this.message, value);
        }

        #endregion
        #region 텍스트 - Text

        /// <summary>
        /// 텍스트
        /// </summary>
        public string Text
        {
            get => this.text;
            set => Set(ref this.text, value);
        }

        #endregion

        //////////////////////////////////////////////////////////////////////////////////////////////////// Constructor
        ////////////////////////////////////////////////////////////////////////////////////////// Public

        #region 생성자 - InputWindow()

        /// <summary>
        /// 생성자
        /// </summary>
        public InputWindow()
        {
            InitializeComponent();

            this.okButton.Click     += okButton_Click;
            this.cancelButton.Click += cancelButton_Click;

            DataContext = this;
        }

        #endregion

        //////////////////////////////////////////////////////////////////////////////////////////////////// Method
        ////////////////////////////////////////////////////////////////////////////////////////// Static
        //////////////////////////////////////////////////////////////////////////////// Public

        #region 표시하기 - Show(message, title, initialValue)

        /// <summary>
        /// 표시하기
        /// </summary>
        /// <param name="message">메시지</param>
        /// <param name="title">제목</param>
        /// <param name="initialValue">초기 값</param>
        /// <returns>문자열</returns>
        public static string Show(string message, string title, string initialValue = null)
        {
            InputWindow window = new InputWindow
            {
                Title   = title,
                Message = message,
                Text    = initialValue
            };

            if(window.ShowDialog() is true)
            {
                return window.Text;
            }

            return null;
        }

        #endregion

        ////////////////////////////////////////////////////////////////////////////////////////// Instance
        //////////////////////////////////////////////////////////////////////////////// Private
        ////////////////////////////////////////////////////////////////////// Event

        #region 확인 버튼 클릭시 처리하기 - okButton_Click(sender, e)

        /// <summary>
        /// 확인 버튼 클릭시 처리하기
        /// </summary>
        /// <param name="sender">이벤트 발생자</param>
        /// <param name="e">이벤트 인자</param>
        private void okButton_Click(object sender, RoutedEventArgs e)
        {
            DialogResult = true;
        }

        #endregion
        #region 취소 버튼 클릭시 처리하기 - cancelButton_Click(sender, e)

        /// <summary>
        /// 취소 버튼 클릭시 처리하기
        /// </summary>
        /// <param name="sender">이벤트 발생자</param>
        /// <param name="e">이벤트 인자</param>
        private void cancelButton_Click(object sender, RoutedEventArgs e)
        {
            DialogResult = false;
        }

        #endregion

        ////////////////////////////////////////////////////////////////////// Function

        #region 설정하기 - Set<TField>(field, newValue, propertyName)

        /// <summary>
        /// 설정하기
        /// </summary>
        /// <typeparam name="TField">필드 타입</typeparam>
        /// <param name="field">필드</param>
        /// <param name="newValue">신규 값</param>
        /// <param name="propertyName">속성명</param>
        /// <returns>처리 결과</returns>
        private bool Set<TField>(ref TField field, TField newValue, [CallerMemberName] string propertyName = null)
        {
            if(EqualityComparer<TField>.Default.Equals(field, newValue))
            {
                return false;
            }

            field = newValue;

            PropertyChanged?.Invoke(this, new PropertyChangedEventArgs(propertyName));

            return true;
        }

        #endregion
    }
}