﻿using Microsoft.Win32;
using System;
using System.Collections.ObjectModel;
using System.ComponentModel;
using System.Runtime.CompilerServices;
using System.Windows;
using System.Windows.Media.Animation;

using TestLibrary;

namespace TestProject
{
    /// <summary>
    /// 메인 윈도우
    /// </summary>
    public partial class MainWindow : Window, INotifyPropertyChanged
    {
        //////////////////////////////////////////////////////////////////////////////////////////////////// Event
        ////////////////////////////////////////////////////////////////////////////////////////// Public

        #region 속성 변경시 이벤트 - PropertyChanged

        /// <summary>
        /// 속성 변경시 이벤트
        /// </summary>
        public event PropertyChangedEventHandler PropertyChanged;

        #endregion

        //////////////////////////////////////////////////////////////////////////////////////////////////// Field
        ////////////////////////////////////////////////////////////////////////////////////////// Private

        #region Field

        /// <summary>
        /// 디폴트 반복 동작 사용 여부
        /// </summary>
        private bool useDefaultRepeatBehavior = true;

        /// <summary>
        /// 무한 반복 여부
        /// </summary>
        private bool repeatForever;

        /// <summary>
        /// 지정 반복 카운트 사용 여부
        /// </summary>
        private bool useSpecificRepeatCount;

        /// <summary>
        /// 반복 카운트
        /// </summary>
        private int repeatCount = 3;

        /// <summary>
        /// 반복 동작
        /// </summary>
        private RepeatBehavior repeatBehavior;

        /// <summary>
        /// 자동 시작 여부
        /// </summary>
        private bool autoStart = true;

        /// <summary>
        /// 다운로드 여부
        /// </summary>
        private bool isDownloading;

        /// <summary>
        /// 다운로드 진행률
        /// </summary>
        private int downloadProgress;

        /// <summary>
        /// 다운로드 진행률 미정 여부
        /// </summary>
        private bool isDownloadProgressIndeterminate;

        /// <summary>
        /// 완료 여부
        /// </summary>
        private bool completed;

        /// <summary>
        /// 이미지 컬렉션
        /// </summary>
        private ObservableCollection<string> imageCollection;

        /// <summary>
        /// 선택 이미지
        /// </summary>
        private string selectedImage;

        /// <summary>
        /// 애니메이터
        /// </summary>
        private Animator animator;

        #endregion

        //////////////////////////////////////////////////////////////////////////////////////////////////// Property
        ////////////////////////////////////////////////////////////////////////////////////////// Public

        #region 디폴트 반복 동작 사용 여부 - UseDefaultRepeatBehavior

        /// <summary>
        /// 디폴트 반복 동작 사용 여부
        /// </summary>
        public bool UseDefaultRepeatBehavior
        {
            get => this.useDefaultRepeatBehavior;
            set
            {
                this.useDefaultRepeatBehavior = value;

                FirePropertyChangedEvent();

                if(value)
                {
                    RepeatBehavior = default;
                }
            }
        }

        #endregion
        #region 무한 반복 여부 - RepeatForever

        /// <summary>
        /// 무한 반복 여부
        /// </summary>
        public bool RepeatForever
        {
            get => this.repeatForever;
            set
            {
                this.repeatForever = value;

                FirePropertyChangedEvent();

                if(value)
                {
                    RepeatBehavior = RepeatBehavior.Forever;
                }
            }
        }

        #endregion
        #region 지정 반복 카운트 사용 여부 - UseSpecificRepeatCount

        /// <summary>
        /// 지정 반복 카운트 사용 여부
        /// </summary>
        public bool UseSpecificRepeatCount
        {
            get => this.useSpecificRepeatCount;
            set
            {
                this.useSpecificRepeatCount = value;

                FirePropertyChangedEvent();

                if(value)
                {
                    RepeatBehavior = new RepeatBehavior(RepeatCount);
                }
            }
        }

        #endregion
        #region 반복 카운트 - RepeatCount

        /// <summary>
        /// 반복 카운트
        /// </summary>
        public int RepeatCount
        {
            get => this.repeatCount;
            set
            {
                this.repeatCount = value;

                FirePropertyChangedEvent();

                if(UseSpecificRepeatCount)
                {
                    RepeatBehavior = new RepeatBehavior(value);
                }
            }
        }

        #endregion
        #region 반복 동작 - RepeatBehavior

        /// <summary>
        /// 반복 동작
        /// </summary>
        public RepeatBehavior RepeatBehavior
        {
            get => this.repeatBehavior;
            set
            {
                this.repeatBehavior = value;

                FirePropertyChangedEvent();

                Completed = false;
            }
        }

        #endregion
        #region 자동 시작 여부 - AutoStart

        /// <summary>
        /// 자동 시작 여부
        /// </summary>
        public bool AutoStart
        {
            get => this.autoStart;
            set
            {
                this.autoStart = value;

                FirePropertyChangedEvent();
            }
        }

        #endregion
        #region 다운로드 여부 - IsDownloading

        /// <summary>
        /// 다운로드 여부
        /// </summary>
        public bool IsDownloading
        {
            get => this.isDownloading;
            set
            {
                this.isDownloading = value;

                FirePropertyChangedEvent();
            }
        }

        #endregion
        #region 다운로드 진행률 - DownloadProgress

        /// <summary>
        /// 다운로드 진행률
        /// </summary>
        public int DownloadProgress
        {
            get => this.downloadProgress;
            set
            {
                this.downloadProgress = value;

                FirePropertyChangedEvent();
            }
        }

        #endregion
        #region 다운로드 진행률 미정 여부 - IsDownloadProgressIndeterminate

        /// <summary>
        /// 다운로드 진행률 미정 여부
        /// </summary>
        public bool IsDownloadProgressIndeterminate
        {
            get => this.isDownloadProgressIndeterminate;
            set
            {
                this.isDownloadProgressIndeterminate = value;

                FirePropertyChangedEvent();
            }
        }

        #endregion
        #region 완료 여부 - Completed

        /// <summary>
        /// 완료 여부
        /// </summary>
        public bool Completed
        {
            get => this.completed;
            set
            {
                this.completed = value;

                FirePropertyChangedEvent();
            }
        }

        #endregion
        #region 이미지 컬렉션 - ImageCollection

        /// <summary>
        /// 이미지 컬렉션
        /// </summary>
        public ObservableCollection<string> ImageCollection
        {
            get => this.imageCollection;
            set
            {
                this.imageCollection = value;

                FirePropertyChangedEvent();
            }
        }

        #endregion
        #region 선택 이미지 - SelectedImage

        /// <summary>
        /// 선택 이미지
        /// </summary>
        public string SelectedImage
        {
            get => this.selectedImage;
            set
            {
                this.selectedImage = value;

                FirePropertyChangedEvent();

                Completed = false;
            }
        }

        #endregion

        //////////////////////////////////////////////////////////////////////////////////////////////////// Constructor
        ////////////////////////////////////////////////////////////////////////////////////////// Public

        #region 생성자 - MainWindow()

        /// <summary>
        /// 생성자
        /// </summary>
        public MainWindow()
        {
            InitializeComponent();

            this.imageCollection = new ObservableCollection<string>
            {
                "IMAGE/working.gif",
                "IMAGE/earth.gif",
                "IMAGE/radar.gif",
                "IMAGE/bomb.gif",
                "IMAGE/bomb-once.gif",
                "IMAGE/nonanimated.gif",
                "IMAGE/monster.gif",
                "pack://Application:,,,/IMAGE/siteoforigin.gif",
                "IMAGE/partialfirstframe.gif",
                "IMAGE/newton-cradle.gif",
                "IMAGE/not-a-gif.png",
                "IMAGE/optimized-full-code-table.gif",
                "IMAGE/corrupt-frame-size.gif",
                "http://i.imgur.com/rCK6xzh.gif",
                "http://media.giphy.com/media/zW2pe7UscHiq4/giphy.gif",
                "http://media.giphy.com/media/nWn6ko2ygIeEU/giphy.gif"
            };

            this.openFileButton.Click += openFileButton_Click;
            this.openURLButton.Click  += openURLButton_Click;
            this.clearButton.Click    += clearButton_Click;
            this.gcButton.Click       += gcButton_Click;
            this.playButton.Click     += playButton_Click;
            this.pauseButton.Click    += pauseButton_Click;
            this.rewindButton.Click   += rewindButton_Click;

            DataContext = this;
        }

        #endregion

        //////////////////////////////////////////////////////////////////////////////////////////////////// Method
        ////////////////////////////////////////////////////////////////////////////////////////// Protected
        //////////////////////////////////////////////////////////////////////////////// Function

        #region 속성 변경시 이벤트 발생시키기 - FirePropertyChangedEvent(propertyName)

        /// <summary>
        /// 속성 변경시 이벤트 발생시키기
        /// </summary>
        /// <param name="propertyName">속성명</param>
        protected virtual void FirePropertyChangedEvent([CallerMemberName] string propertyName = null)
        {
            PropertyChanged?.Invoke(this, new PropertyChangedEventArgs(propertyName));
        }

        #endregion

        ////////////////////////////////////////////////////////////////////////////////////////// Private
        //////////////////////////////////////////////////////////////////////////////// Event

        #region 애니메이션 동작 로드시 처리하기 - AnimationBehavior_Loaded(sender, e)

        /// <summary>
        /// 애니메이션 동작 로드시 처리하기
        /// </summary>
        /// <param name="sender">이벤트 발생자</param>
        /// <param name="e">이벤트 인자</param>
        private void AnimationBehavior_Loaded(object sender, RoutedEventArgs e)
        {
            IsDownloading = false;

            if(this.animator != null)
            {
                this.animator.CurrentFrameChanged -= Animator_CurrentFrameChanged;
            }

            this.animator = AnimationBehavior.GetAnimator(this.image);

            if(this.animator != null)
            {
                this.animator.CurrentFrameChanged += Animator_CurrentFrameChanged;

                positionSlider.Value   = 0;
                positionSlider.Maximum = this.animator.FrameCount - 1;

                SetPlayControlEnabled(this.animator.IsPaused || this.animator.IsCompleted);
            }
        }

        #endregion
        #region 애니메이션 동작 에러시 처리하기 - AnimationBehavior_Error(d, e)

        /// <summary>
        /// 애니메이션 동작 에러시 처리하기
        /// </summary>
        /// <param name="d">의존 객체</param>
        /// <param name="e">이벤트 인자</param>
        private void AnimationBehavior_Error(DependencyObject d, AnimationErrorEventArgs e)
        {
            if(e.ErrorKind == AnimationErrorKind.Loading)
            {
                IsDownloading = false;
            }

            MessageBox.Show($"에러 발생 ({e.ErrorKind}) : {e.Exception}");
        }

        #endregion
        #region 애니메이션 동작 다운로드 진행률 처리하기 - AnimationBehavior_DownloadProgress(d, e)

        /// <summary>
        /// 애니메이션 동작 다운로드 진행률 처리하기
        /// </summary>
        /// <param name="d">의존 객체</param>
        /// <param name="e">이벤트 인자</param>
        private void AnimationBehavior_DownloadProgress(DependencyObject d, DownloadProgressEventArgs e)
        {
            IsDownloading = true;

            if(e.Progress >= 0)
            {
                DownloadProgress = e.Progress;

                IsDownloadProgressIndeterminate = false;
            }
            else
            {
                IsDownloadProgressIndeterminate = true;
            }
        }

        #endregion
        #region 애니메이션 동작 애니메이션 완료시 처리하기 - AnimationBehavior_AnimationCompleted(sender, e)

        /// <summary>
        /// 애니메이션 동작 애니메이션 완료시 처리하기
        /// </summary>
        /// <param name="sender">이벤트 발생자</param>
        /// <param name="e">이벤트 인자</param>
        private void AnimationBehavior_AnimationCompleted(object sender, AnimationCompletedEventArgs e)
        {
            Completed = true;

            if(this.animator != null)
            {
                SetPlayControlEnabled(this.animator.IsPaused || this.animator.IsCompleted);
            }
        }

        #endregion

        #region 애니메이터 현재 프레임 변경시 처리하기 - Animator_CurrentFrameChanged(sender, e)

        /// <summary>
        /// 애니메이터 현재 프레임 변경시 처리하기
        /// </summary>
        /// <param name="sender">이벤트 발생자</param>
        /// <param name="e">이벤트 인자</param>
        private void Animator_CurrentFrameChanged(object sender, EventArgs e)
        {
            if(this.animator != null)
            {
                this.positionSlider.Value = this.animator.CurrentFrameIndex;
            }
        }

        #endregion

        #region 파일 열기 버튼 클릭시 처리하기 - openFileButton_Click(sender, e)

        /// <summary>
        /// 파일 열기 버튼 클릭시 처리하기
        /// </summary>
        /// <param name="sender">이벤트 발생자</param>
        /// <param name="e">이벤트 인자</param>
        private void openFileButton_Click(object sender, RoutedEventArgs e)
        {
            OpenFileDialog openFileDialog = new OpenFileDialog {Filter = "GIF 이미지|*.gif"};

            if(openFileDialog.ShowDialog() == true)
            {
                ImageCollection.Add(openFileDialog.FileName);

                SelectedImage = openFileDialog.FileName;
            }
        }

        #endregion
        #region URL 열기 버튼 클릭시 처리하기 - openURLButton_Click(sender, e)

        /// <summary>
        /// URL 열기 버튼 클릭시 처리하기
        /// </summary>
        /// <param name="sender">이벤트 발생자</param>
        /// <param name="e">이벤트 인자</param>
        private void openURLButton_Click(object sender, RoutedEventArgs e)
        {
            string url = InputWindow.Show("로드할 이미지 URL을 입력해 주시기 바랍니다.", "URL 입력");

            if(!string.IsNullOrEmpty(url))
            {
                ImageCollection.Add(url);

                SelectedImage = url;
            }
        }

        #endregion
        #region 지우기 버튼 클릭시 처리하기 - clearButton_Click(sender, e)

        /// <summary>
        /// 지우기 버튼 클릭시 처리하기
        /// </summary>
        /// <param name="sender">이벤트 발생자</param>
        /// <param name="e">이벤트 인자</param>
        private void clearButton_Click(object sender, RoutedEventArgs e)
        {
            SelectedImage = null;
        }

        #endregion
        #region GC 버튼 클릭시 처리하기 - gcButton_Click(sender, e)

        /// <summary>
        /// GC 버튼 클릭시 처리하기
        /// </summary>
        /// <param name="sender">이벤트 발생자</param>
        /// <param name="e">이벤트 인자</param>
        private void gcButton_Click(object sender, RoutedEventArgs e)
        {
            GC.Collect();

            GC.WaitForPendingFinalizers();
        }

        #endregion
        #region 재생 버튼 클릭시 처리하기 - playButton_Click(sender, e)

        /// <summary>
        /// 재생 버튼 클릭시 처리하기
        /// </summary>
        /// <param name="sender">이벤트 발생자</param>
        /// <param name="e">이벤트 인자</param>
        private void playButton_Click(object sender, RoutedEventArgs e)
        {
            this.animator?.Play();

            Completed = false;

            SetPlayControlEnabled(false);
        }

        #endregion
        #region 중지 버튼 클릭시 처리하기 - pauseButton_Click(sender, e)

        /// <summary>
        /// 중지 버튼 클릭시 처리하기
        /// </summary>
        /// <param name="sender">이벤트 발생자</param>
        /// <param name="e">이벤트 인자</param>
        private void pauseButton_Click(object sender, RoutedEventArgs e)
        {
            this.animator?.Pause();

            SetPlayControlEnabled(true);
        }

        #endregion
        #region 되감기 버튼 클릭시 처리하기 - rewindButton_Click(sender, e)

        /// <summary>
        /// 되감기 버튼 클릭시 처리하기
        /// </summary>
        /// <param name="sender">이벤트 발생자</param>
        /// <param name="e">이벤트 인자</param>
        private void rewindButton_Click(object sender, RoutedEventArgs e)
        {
            if(this.animator == null)
            {
                return;
            }

            this.animator.Rewind();

            SetPlayControlEnabled(this.animator.IsPaused || this.animator.IsCompleted);

            Completed = false;
        }

        #endregion

        //////////////////////////////////////////////////////////////////////////////// Function

        #region 재생 컨트롤 이용 가능 여부 설정하기 - SetPlayControlEnabled(isPaused)

        /// <summary>
        /// 재생 컨트롤 이용 가능 여부 설정하기
        /// </summary>
        /// <param name="isPaused">중지 여부</param>
        private void SetPlayControlEnabled(bool isPaused)
        {
            this.pauseButton.IsEnabled  = !isPaused;
            this.playButton.IsEnabled   = isPaused;
            this.rewindButton.IsEnabled = true;
        }

        #endregion
    }
}